
from __future__ import annotations

import json
import traceback
from dataclasses import dataclass, asdict
from typing import List, Dict
from natsort import natsorted

import os
try:
    from importlib import resources as impresources
    BASE_FOLDER = impresources.files(__name__)
except:
    BASE_FOLDER = os.path.split(__file__)[0]

@dataclass
class Conversation():
    messages: List[Message] = None

    @classmethod
    def from_dict(cls, d):
        if 'messages' in d:
            messages = [Message.from_dict(e) for e in d['messages']]
        else:
            messages = None
        return cls(messages=messages)

    def validate(self):
        assert isinstance(self.messages, list) and \
            all(isinstance(message, Message) for message in self.messages) and \
            all(message.validate() for message in self.messages)
        return True

@dataclass
class Message():
    role: str = None
    text: str = None
    error: str = None

    @classmethod
    def from_dict(cls, d):
        m = cls(role=d.get('role', None), text=d.get('text', None), error=d.get('error', None))
        return m

    def validate(self):
        assert self.role in ['System', 'User', 'Imputation'] or self.role.startswith('Model.')
        assert (isinstance(self.text, str) and self.error in [None, 'LENGTH_CURRENT']) or \
               (self.text is None and self.error == 'LENGTH_PREVIOUS')
        if self.error is not None:
            assert self.role.startswith('Model.')
        return True

@dataclass
class FWSAnnotation():
    text: str = None
    start_idx: int = None
    end_idx: int = None
    author: str = None

    @classmethod
    def from_dict(cls, d):
        return cls(text=d.get('text', None), start_idx=d.get('start_idx', None), 
                   end_idx=d.get('end_idx', None), author=d.get('author', None))

    def wrong_step(self):
        if self.start_idx is None:
            return 'No mistake'
        else:
            return self.text[self.start_idx : self.end_idx]

    def validate(self):
        assert isinstance(self.text, str)
        assert self.start_idx is None or isinstance(self.start_idx, int)
        assert self.end_idx is None or isinstance(self.end_idx, int)
        assert (self.start_idx is None) == (self.end_idx is None)
        assert isinstance(self.author, str)
        return True

@dataclass
class Concept():
    identifier: str = None
    name: str = None
    _raw_text: str = None
    _text: str = None
    category: str = None
    parent: Concept = None

    @classmethod
    def from_dict(cls, d):
        if d is None:
            return None
        parent = cls.from_dict(d['parent']) if 'parent' in d else None
        c = cls(identifier=d.get('identifier', None), name=d.get('name', None), 
                   _raw_text=d.get('_raw_text', None), category=d.get('category', None), 
                   parent=parent)
        if c._raw_text is not None:
            c.raw_text = c._raw_text
        return c

    @property
    def raw_text(self):
        return self._raw_text

    @raw_text.setter
    def raw_text(self, v):
        assert v.count('@@') % 2 == 0, f'Incorrect text format: {v}'
        self._raw_text = v
        self._text = v.replace('@@', '')

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, v):
        raise NotImplementedError('"text" should not be set directly. Set "raw_text" instead.')

    def get_root(self):
        return self.get_line()[-1]

    def get_line(self):
        cur = self
        line = [cur]
        while cur.parent is not None:
            cur = cur.parent
            line.append(cur)
        return line

    def validate(self):
        assert isinstance(self.identifier, str) and self.identifier.startswith('C_')
        assert isinstance(self.text, str)
        assert isinstance(self.category, str)
        assert self.name is None or isinstance(self.name, str)
        assert self.parent is None or (isinstance(self.parent, Concept) and self.parent.validate())
        return True

@dataclass
class Hint():
    identifier: str = None
    _raw_text: str = None
    _text: str = None

    @classmethod
    def from_dict(cls, d):
        h = cls(identifier=d.get('identifier', None), _raw_text=d.get('_raw_text', None))
        if h._raw_text is not None:
            h.raw_text = h._raw_text
        return h

    @property
    def raw_text(self):
        return self._raw_text

    @raw_text.setter
    def raw_text(self, v):
        assert v.count('@@') % 2 == 0, f'Incorrect text format: {v}'
        self._raw_text = v
        self._text = v.replace('@@', '')

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, v):
        raise NotImplementedError('"text" should not be set directly. Set "raw_text" instead.')

    def validate(self):
        assert isinstance(self.identifier, str) and self.identifier.startswith('H_')
        assert isinstance(self.text, str)
        return True

@dataclass
class SolutionStep():
    _raw_text: str = None
    _text: str = None
    ch_idxs: List[int] = None

    @classmethod
    def from_dict(cls, d):
        st = cls(_raw_text=d.get('_raw_text', None), ch_idxs=d.get('ch_idxs', None))
        if st._raw_text is not None:
            st.raw_text = st._raw_text
        return st

    @property
    def raw_text(self):
        return self._raw_text

    @raw_text.setter
    def raw_text(self, v):
        assert v.count('@@') % 2 == 0, f'Incorrect text format: {v}'
        self._raw_text = v
        self._text = v.replace('@@', '')

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, v):
        raise NotImplementedError('"text" should not be set directly. Set "raw_text" instead.')

    def validate(self):
        assert isinstance(self.text, str)
        assert isinstance(self.ch_idxs, list) and all(isinstance(idx, int) for idx in self.ch_idxs)
        return True

@dataclass
class Solution():
    steps: List[SolutionStep] = None

    @classmethod
    def from_dict(cls, d):
        if 'steps' in d:
            steps = [SolutionStep.from_dict(e) for e in d['steps']]
        else:
            steps = None
        return cls(steps=steps)

    def validate(self):
        assert isinstance(self.steps, list) and \
        all(isinstance(step, SolutionStep) for step in self.steps) and \
        all(step.validate() for step in self.steps)
        return True

@dataclass
class Problem():
    identifier: str = None
    _raw_text: str = None
    _text: str = None
    category: str = None
    solution: Solution = None
    _raw_answer: str = None
    _answer: str = None
    ch_list: List[str] = None
    fws_annotations: Dict[str, FWSAnnotation] = None
    conversations: Dict[str, Conversation] = None

    @classmethod
    def from_dict(cls, d):
        solution = Solution.from_dict(d['solution']) if 'solution' in d else None
        if 'conversations' in d and d['conversations'] is not None:
            conversations = dict()
            for k, v in d['conversations'].items():
                conversations[k] = Conversation.from_dict(v)
        else:
            conversations = None
        if 'fws_annotations' in d and d['fws_annotations'] is not None:
            fws_annotations = dict()
            for k, v in d['fws_annotations'].items():
                fws_annotations[k] = FWSAnnotation.from_dict(v)
        else:
            fws_annotations = None
        p = cls(identifier=d.get('identifier', None), _raw_text=d.get('_raw_text', None), 
                category=d.get('category', None), solution=solution, 
                _raw_answer=d.get('_raw_answer', None), ch_list=d.get('ch_list', None), 
                fws_annotations=fws_annotations, conversations=conversations)
        if p._raw_text is not None:
            p.raw_text = p._raw_text
        if p._raw_answer is not None:
            p.raw_answer = p._raw_answer
        return p

    @property
    def raw_text(self):
        return self._raw_text

    @raw_text.setter
    def raw_text(self, v):
        assert v.count('@@') % 2 == 0, f'Incorrect text format: {v}'
        self._raw_text = v
        self._text = v.replace('@@', '')

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, v):
        raise NotImplementedError('"text" should not be set directly. Set "raw_text" instead.')

    @property
    def raw_answer(self):
        return self._raw_answer

    @raw_answer.setter
    def raw_answer(self, v):
        assert v.count('@@') % 2 == 0, f'Incorrect answer format: {v}'
        self._raw_answer = v
        self._answer = v.replace('@@', '')

    @property
    def answer(self):
        return self._answer

    @answer.setter
    def answer(self, v):
        raise NotImplementedError('"answer" should not be set directly. Set "raw_answer" instead.')

    def unique_ch_list(self):
        s = set()
        uniq_list = []
        for e in self.ch_list:
            if e.identifier not in s:
                s.add(e.identifier)
                uniq_list.append(e)
        return uniq_list

    def validate(self):
        assert isinstance(self.identifier, str) and self.identifier.startswith('P_')
        assert isinstance(self.text, str)
        assert isinstance(self.category, str)
        assert isinstance(self.solution, Solution) and self.solution.validate()
        assert isinstance(self.answer, str)
        assert isinstance(self.ch_list, list) and all(isinstance(ch, str) for ch in self.ch_list)
        assert isinstance(self.fws_annotations, dict) and \
               all(isinstance(k, str) for k in self.fws_annotations.keys()) and \
               all(isinstance(v, FWSAnnotation) for v in self.fws_annotations.values()) and \
               all(v.validate() for v in self.fws_annotations.values())
        assert isinstance(self.conversations, dict) and \
               all(isinstance(k, str) for k in self.conversations.keys()) and \
               all(k.count('|')==1 for k in self.conversations.keys()) and \
               all(isinstance(v, Conversation) for v in self.conversations.values()) and \
               all(v.validate() for v in self.conversations.values())
        return True

@dataclass
class Dataset():
    problems: Dict[str, Problem] = None
    concepts: Dict[str, Concept] = None
    hints: Dict[str, Hint] = None

    @classmethod
    def from_json(cls, fn):
        with open(fn) as f:
            d = json.load(f)
        dataset = cls.from_dict(d)
        try:
            assert dataset.validate()
        except Exception as e:
            traceback.print_exc()
            print('Dataset validation fails')
        return dataset

    @classmethod
    def from_dict(cls, d):
        if 'problems' in d:
            problems = dict()
            for k, v in d['problems'].items():
                problems[k] = Problem.from_dict(v)
        else:
            problems = None
        if 'concepts' in d:
            concepts = dict()
            for k, v in d['concepts'].items():
                concepts[k] = Concept.from_dict(v)
        else:
            concepts = None
        if 'hints' in d:
            hints = dict()
            for k, v in d['hints'].items():
                hints[k] = Hint.from_dict(v)
        else:
            hints = None
        return cls(problems=problems, concepts=concepts, hints=hints)

    def __getitem__(self, identifier):
        if identifier[0] == 'P':
            return self.problems[identifier]
        elif identifier[0] == 'C':
            return self.concepts[identifier]
        elif identifier[0] == 'H':
            return self.hints[identifier]
        else:
            raise Exception(f'Unrecognized identifier {identifier}')

    def prune_ch(self, verbose=False):
        used_chs = set()
        for p in self.problems.values():
            for ch in p.ch_list:
                used_chs.add(ch.identifier)
        if verbose:
            print((f'We have {len(self.concepts)} concepts and {len(self.hints)} hints'))
            k_pruned = [k for k, v in self.concepts.items() if k not in used_chs]
            h_pruned = [k for k, v in self.hints.items() if k not in used_chs]
            print('Pruning concepts:', k_pruned)
            print('Pruning hints:', h_pruned)
        self.concepts = {k: v for k, v in self.concepts.items() if k in used_chs}
        self.hints = {k: v for k, v in self.hints.items() if k in used_chs}
        if verbose:
            print((f'We now have {len(self.concepts)} concepts and {len(self.hints)} hints'))

    def to_json(self, fn):
        with open(fn, 'w') as f:
            json.dump(asdict(self), f, indent=2)

    def validate(self):
        for p in self.problems.values():
            p.validate()
        for c in self.concepts.values():
            c.validate()
        for h in self.hints.values():
            h.validate()
        return True

    def sort(self):
        self.problems = {k: self.problems[k] for k in natsorted(self.problems.keys())}
        self.concepts = {k: self.concepts[k] for k in natsorted(self.concepts.keys())}
        self.hints = {k: self.hints[k] for k in natsorted(self.hints.keys())}

def load(fn):
    if not fn.endswith('.json'):
        fn = os.path.join(BASE_FOLDER, f'dataset_files/{fn}.json')
    with open(fn) as f:
        data = json.load(f)
    dataset = Dataset.from_dict(data)
    return dataset

from .prompt import PromptGenerator




import random
from copy import deepcopy as copy
from functools import partial
from . import *

sys_prompt = 'You are an expert on mathematics.'

prompt_main = '\n\n'.join([
    'Solve the following problem. Make sure to show your work before giving the final answer.', 
    '{problem.text}'
])

prompt_ch_list = '\n\n'.join([
    prompt_main, 
    'You may find the following information useful:',
    '{ch_list_str}'
])

prompt_name = 'Please explain the following concept: {concept.name}.'

prompt_example = '\n\n'.join([
    'Please give an example that applies the following concept:', 
    '{concept.text}'
])

prompt_name_apply_has_remaining_ch = '\n\n'.join([
    prompt_main, 
    'Besides the concept{plural_s} above, you may also find the following information useful:', 
    '{rem_ch_list_str}'
])

prompt_name_apply_no_remaining_ch = '\n\n'.join([
    prompt_main, 
    'You may find the above concept{plural_s} helpful.'
])

prompt_problem_single = '\n\n'.join([
    'First, study the following problem and its solution as they may contain helpful information.', 
    '{sample_problem}'
])

prompt_problem_multiple = '\n\n'.join([
    'First, study the following problems and their solutions as they may contain helpful information.',
    '{sample_problems}'
])

prompt_problem_query = ('\n\nWith what you have learned, solve the following problem. '
                 'Make sure to show your work before giving the final answer.\n\n{problem.text}')

prompt_problem_remaining = ('\n\nIn addition, you may also find the following information helpful:'
                '\n\n{rem_ch_list_str}')


prompt_partial_soln = '\n\n'.join([
    prompt_main, 
    'Below is a partial solution to the problem that may be helpful:', 
    '{soln_list_str}'
])

prompt_summarizer = ('Now, summarize the answer above in one sentence, '
                     'without any intermediate steps or explanations.')

def plural_s(i):
    assert i != 0 and isinstance(i, int)
    return '' if i == 1 else 's'

class PromptGenerator():
    def __init__(self, dataset):
        self.dataset = dataset

    def show_solution(self, pid):
        p = self.dataset[pid]
        result = f'Problem: {p.text}\nStep-wise solution:\n'
        result += self.enumerate_solution(p.solution.steps)
        return result

    def enumerate_solution(self, steps):
        return '\n'.join([f'{idx+1}. {step.text}' for idx, step in enumerate(steps)])

    def enumerate_ch(self, ch_list):
        assert len(ch_list) != 0
        return '\n'.join([f'{idx+1}. {self.dataset[ch].text}' for idx, ch in enumerate(ch_list)])

    def problem_only(self, problem):
        return sys_prompt, [prompt_main.format(problem=problem)], []

    def problem_with_ch_list(self, problem, ch_list):
        if len(ch_list) == 0:
            return self.problem_only(problem)
        ch_list_str = self.enumerate_ch(ch_list)
        return sys_prompt, [prompt_ch_list.format(problem=problem, ch_list_str=ch_list_str)], []

    def no_c(self, problem, with_h):
        if with_h:
            ch_list = [ch for ch in problem.ch_list if ch.startswith('H')]
        else:
            ch_list = []
        return self.problem_with_ch_list(problem, ch_list)

    def direct_c(self, problem, with_h):
        if with_h:
            ch_list = problem.ch_list
        else:
            ch_list = [ch for ch in problem.ch_list if ch.startswith('C')]
        return self.problem_with_ch_list(problem, ch_list)

    def name_c(self, problem, with_h):
        prompts = []
        remaining_chs = []
        for ch in problem.ch_list:
            if ch.startswith('C'):
                if self.dataset[ch].name is not None:
                    prompts.append(prompt_name.format(concept=self.dataset[ch]))
                else:
                    remaining_chs.append(ch)
            else:
                if with_h:
                    remaining_chs.append(ch)
        if len(prompts) == 0:
            return self.direct_c(problem, with_h)
        if len(remaining_chs) != 0:
            query_prompt = prompt_name_apply_has_remaining_ch.format(
                problem=problem, 
                plural_s=plural_s(len(prompts)), 
                rem_ch_list_str=self.enumerate_ch(remaining_chs))
        else:
            query_prompt = prompt_name_apply_no_remaining_ch.format(
                problem=problem, 
                plural_s=plural_s(len(prompts)))
        prompts.append(query_prompt)
        return sys_prompt, prompts, []

    def example_c(self, problem, with_h):
        prompts = []
        for ch in problem.ch_list:
            if ch.startswith('C'):
                prompts.append(prompt_example.format(concept=self.dataset[ch]))
        if len(prompts) == 0:
            return self.direct_c(problem, with_h)
        if with_h:
            hs = [ch for ch in problem.ch_list if ch.startswith('H')]
        else:
            hs = []
        if len(hs) != 0:
            query_prompt = prompt_name_apply_has_remaining_ch.format(
                problem=problem, 
                plural_s=plural_s(len(prompts)), 
                rem_ch_list_str=self.enumerate_ch(hs))
        else:
            query_prompt = prompt_name_apply_no_remaining_ch.format(
                problem=problem, 
                plural_s=plural_s(len(prompts)))
        prompts.append(query_prompt)
        return sys_prompt, prompts, []

    def root_c(self, problem, with_h):
        ch_instances = [self.dataset[ch] for ch in problem.ch_list]
        if with_h:
            ch_list = [ch.get_root().identifier if isinstance(ch, Concept) else ch.identifier 
                for ch in ch_instances]
        else:
            ch_list = [ch.get_root().identifier for ch in ch_instances if isinstance(ch, Concept)]
        return self.problem_with_ch_list(problem, ch_list)

    def build_c2p_dict(self):
        try:
            self.c2p
        except AttributeError:
            self.c2p = dict()
            for pid, p in self.dataset.problems.items():
                for ch in p.ch_list:
                    if ch.startswith('H'):
                        continue
                    if ch not in self.c2p:
                        self.c2p[ch] = []
                    self.c2p[ch].append(pid)
            for k, v in self.c2p.items():
                self.c2p[k] = sorted(v)

    def problem_c(self, problem, with_h):
        self.build_c2p_dict()
        remaining_chs = []
        probs = []
        for ch in problem.ch_list:
            if ch.startswith('C'):
                pids = [pid for pid in self.c2p[ch] if pid != problem.identifier]
                if len(pids) == 0:
                    remaining_chs.append(ch)
                else:
                    probs.append(random.choice(pids))
            else:
                if with_h:
                    remaining_chs.append(ch)
        probs_uniq = []
        for p in probs:
            if p not in probs_uniq:
                probs_uniq.append(p)
        probs = probs_uniq
        if len(probs) == 0:
            return self.direct_c(problem, with_h)
        prompt_lst = []
        if len(probs) == 1:
            prompt_lst.append(prompt_problem_single.format(
                sample_problem=self.show_solution(probs[0])))
        else:
            samples = '\n\n'.join([self.show_solution(prob) for prob in probs])
            prompt_lst.append(prompt_problem_multiple.format(
                sample_problems=samples))
        prompt_lst.append(prompt_problem_query.format(problem=problem))
        if len(remaining_chs) != 0:
            prompt_lst.append(prompt_problem_remaining.format(
                rem_ch_list_str=self.enumerate_ch(remaining_chs)))
        return sys_prompt, [''.join(prompt_lst)], []

    def build_c_cats_dict(self):
        try:
            self.c_cats
        except AttributeError:
            self.c_cats = dict()
            for cid, c in self.dataset.concepts.items():
                cat = c.category
                if cat not in self.c_cats:
                    self.c_cats[cat] = []
                self.c_cats[cat].append(cid)
            for c, v in self.c_cats.items():
                self.c_cats[c] = sorted(v)

    def misleading_c(self, problem, with_h):
        self.build_c_cats_dict()
        avoid_chs = set()
        for ch in problem.ch_list:
            if ch.startswith('C'):
                avoid_chs.update([c.identifier for c in self.dataset[ch].get_line()])
        new_chs = []
        for ch in problem.ch_list:
            if ch.startswith('H'):
                if with_h:
                    new_chs.append(ch)
            else:
                candidates = [c for c in self.c_cats[self.dataset[ch].category] if c not in avoid_chs]
                if len(candidates) == 0:
                    new_chs.append(ch)
                else:
                    new_chs.append(random.choice(candidates))
        return self.problem_with_ch_list(problem, new_chs)

    def partial_solution(self, problem, soln_frac):
        assert 0 <= soln_frac <= 1
        frac_num = (len(problem.solution.steps) - 1) * soln_frac
        whole = int(frac_num)
        if random.random() < (frac_num - whole):
            whole += 1
        if whole == 0:
            return self.problem_only(problem)
        soln_str = self.enumerate_solution(problem.solution.steps[:whole])
        prompt = prompt_partial_soln.format(
            problem=problem, 
            soln_list_str=soln_str)
        return sys_prompt, [prompt], []

    def few_shot(self, problem, num_shots):
        pids = [pid for pid in self.dataset.problems if pid != problem.identifier]
        pids = random.sample(pids, num_shots)
        prompts = []
        imputed_outputs = []
        for pid in pids:
            shot_prob = self.dataset.problems[pid]
            shot_input = prompt_main.format(problem=shot_prob)
            shot_output = 'Step-wise solution:\n'+self.enumerate_solution(shot_prob.solution.steps)
            prompts.append(shot_input)
            imputed_outputs.append(shot_output)
        p_input = prompt_main.format(problem=problem)
        prompts.append(p_input)
        return sys_prompt, prompts, imputed_outputs

    def get_all_prompts(self):
        return {'0-Shot': self.problem_only, 
                '5-Shot': partial(self.few_shot, num_shots=5), 
                '1/3 Soln': partial(self.partial_solution, soln_frac=1/3),
                '2/3 Soln': partial(self.partial_solution, soln_frac=2/3),
                'No C w/o H': partial(self.no_c, with_h=False), 
                'No C w/ H': partial(self.no_c, with_h=True), 
                'Direct C w/o H': partial(self.direct_c, with_h=False), 
                'Direct C w/ H': partial(self.direct_c, with_h=True), 
                'Name C w/o H': partial(self.name_c, with_h=False), 
                'Name C w/ H': partial(self.name_c, with_h=True), 
                'Example C w/o H': partial(self.example_c, with_h=False), 
                'Example C w/ H': partial(self.example_c, with_h=True), 
                'Root C w/o H': partial(self.root_c, with_h=False), 
                'Root C w/ H': partial(self.root_c, with_h=True), 
                'Problem C w/o H': partial(self.problem_c, with_h=False), 
                'Problem C w/ H': partial(self.problem_c, with_h=True), 
                'Misleading C w/o H': partial(self.misleading_c, with_h=False), 
                'Misleading C w/ H': partial(self.misleading_c, with_h=True)
               }

    def get_all_prompt_names(self):
        return list(self.get_all_prompts().keys())

    def construct_prompt(self, prompt_name, problem):
        prompts = self.get_all_prompts()
        if prompt_name not in prompts:
            raise ValueError(f'Invalid prompt name: {prompt_name}. ' + \
                f'The list of choices is: {list(prompts.keys())}')
        func = prompts[prompt_name]
        return func(problem)



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name = 'champ-dataset',
    version = '1.0.0',
    author = 'Yilun Zhou',
    author_email = 'zyilun94@gmail.com',
    description = 'Concept and Hint-Augmented Math Dataset (CHAMP)',
    license = 'MIT',
    keywords = ['natural language processing', 'large language model'],
    url = 'https://yujunmao1.github.io/CHAMP/',
    packages=['champ_dataset'],
    long_description = long_description,
    long_description_content_type='text/markdown', 
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research', 
        'Topic :: Scientific/Engineering :: Artificial Intelligence', 
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python', 
    ],
    install_requires = ['natsort'], 
    include_package_data = True, 
)